#!/usr/bin/env python

import unittest
import tempfile
import shutil
import os
from nebulizer.core import Credentials

class TestCredentials(unittest.TestCase):
    """
    Tests for the 'Credentials' class

    """
    def setUp(self):
        # Create temp working dir
        self.tmpdir = tempfile.mkdtemp(suffix='TestCredentials')

    def tearDown(self):
        # Remove the temporary test directory
        shutil.rmtree(self.tmpdir)

    def _make_key_file(self):
        # Create a key file for testing
        tmp_key_file = os.path.join(self.tmpdir,'.nebulizer')
        with open(tmp_key_file,'w') as fp:
            fp.write("""# .nebulizer
production\thttp://prod.example.org\t37b6444b8c62a137ab306242
devel\thttp://devel.example.org\t137ab30624237b6444b8c62a
local\thttp://127.0.0.1:8080\tb8c62624237b6444137ab30
""")
        return tmp_key_file

    def test_list_keys(self):
        """
        Credentials.list_keys: lists aliases
        """
        tmp_key_file = self._make_key_file()
        credentials = Credentials(key_file=tmp_key_file)
        self.assertEqual(credentials.list_keys(),
                         ['production',
                          'devel',
                          'local'])

    def test_fetch_key(self):
        """
        Credentials.fetch_key: fetches correct data from key file
        """
        tmp_key_file = self._make_key_file()
        credentials = Credentials(key_file=tmp_key_file)
        self.assertEqual(credentials.fetch_key('production'),
                         ('http://prod.example.org',
                          '37b6444b8c62a137ab306242'))
        self.assertEqual(credentials.fetch_key('devel'),
                         ('http://devel.example.org',
                          '137ab30624237b6444b8c62a'))
        self.assertEqual(credentials.fetch_key('local'),
                         ('http://127.0.0.1:8080',
                          'b8c62624237b6444137ab30'))
        self.assertEqual(credentials.fetch_key('http://devel.example.org'),
                         ('http://devel.example.org',
                          '137ab30624237b6444b8c62a'))
        self.assertRaises(KeyError,credentials.fetch_key,'nonexistent')

    def test_store_key(self):
        """
        Credentials.store_key: appends new key to key file
        """
        tmp_key_file = self._make_key_file()
        credentials = Credentials(key_file=tmp_key_file)
        self.assertEqual(credentials.list_keys(),
                         ['production',
                          'devel',
                          'local'])
        credentials.store_key('beta-staging',
                              'http://beta-staging.example.org',
                              '137ab30624237b6444b8c62a')
        self.assertEqual(credentials.list_keys(),
                         ['production',
                          'devel',
                          'local',
                          'beta-staging'])
        self.assertEqual(credentials.fetch_key('production'),
                         ('http://prod.example.org',
                          '37b6444b8c62a137ab306242'))
        self.assertEqual(credentials.fetch_key('devel'),
                         ('http://devel.example.org',
                          '137ab30624237b6444b8c62a'))
        self.assertEqual(credentials.fetch_key('local'),
                         ('http://127.0.0.1:8080',
                          'b8c62624237b6444137ab30'))
        self.assertEqual(credentials.fetch_key('beta-staging'),
                         ('http://beta-staging.example.org',
                          '137ab30624237b6444b8c62a'))

    def test_remove_key(self):
        """
        Credentials.remove_key: removes key from key file
        """
        tmp_key_file = self._make_key_file()
        credentials = Credentials(key_file=tmp_key_file)
        self.assertEqual(credentials.list_keys(),
                         ['production',
                          'devel',
                          'local'])
        credentials.remove_key('devel')
        self.assertEqual(credentials.list_keys(),
                         ['production',
                          'local'])
        self.assertEqual(credentials.fetch_key('production'),
                         ('http://prod.example.org',
                          '37b6444b8c62a137ab306242'))
        self.assertEqual(credentials.fetch_key('local'),
                         ('http://127.0.0.1:8080',
                          'b8c62624237b6444137ab30'))
        credentials.remove_key('nonexistent')
        self.assertEqual(credentials.list_keys(),
                         ['production',
                          'local'])
        self.assertEqual(credentials.fetch_key('production'),
                         ('http://prod.example.org',
                          '37b6444b8c62a137ab306242'))
        self.assertEqual(credentials.fetch_key('local'),
                         ('http://127.0.0.1:8080',
                          'b8c62624237b6444137ab30'))

    def test_update_key(self):
        """
        Credentials.update_key: updates details in key file
        """
        tmp_key_file = self._make_key_file()
        credentials = Credentials(key_file=tmp_key_file)
        self.assertEqual(credentials.list_keys(),
                         ['production',
                          'devel',
                          'local'])
        self.assertEqual(credentials.fetch_key('devel'),
                         ('http://devel.example.org',
                          '137ab30624237b6444b8c62a'))
        credentials.update_key('devel')
        self.assertEqual(credentials.fetch_key('devel'),
                         ('http://devel.example.org',
                          '137ab30624237b6444b8c62a'))
        credentials.update_key('devel',
                               new_url='http://devel2.example.org')
        self.assertEqual(credentials.fetch_key('devel'),
                         ('http://devel2.example.org',
                          '137ab30624237b6444b8c62a'))
        credentials.update_key('devel',
                               new_api_key='eadef00d245c3d0135ba9ccae7e77aef')
        self.assertEqual(credentials.fetch_key('devel'),
                         ('http://devel2.example.org',
                          'eadef00d245c3d0135ba9ccae7e77aef'))
        credentials.update_key('devel',
                               new_url='http://devel.example.org',
                               new_api_key='137ab30624237b6444b8c62a')



#!/usr/bin/env python
#
# users: functions for managing users
import logging
import re
import getpass
import fnmatch
from bioblend import galaxy
from mako.template import Template

# Logging
logger = logging.getLogger(__name__)

# Classes

class User:
    """
    Class wrapping extraction of user data

    Provides an interface for accessing data about a user
    in a Galaxy instance, which has been retrieved via a
    a call to the Galaxy API using bioblend.

    """
    def __init__(self,user_data):
        """
        Create a new User instance

        ``user_data`` is a dictionary returned by a
        call to bioblend, for example:

        >>> for user_data in galaxy.users.UserClient(gi).get_users():
        >>>    print User(user_data).name

        """
        self.email = user_data['email']
        self.username = user_data['username']
        self.id = user_data['id']
        self.quota_percent = None
        self.total_disk_usage = None
        self.nice_total_disk_usage = None
        self.is_admin = None
        for attr in user_data.keys():
            try:
                setattr(self,attr,user_data[attr])
            except AttributeError:
                pass

# Functions

def get_users(gi):
    """
    Return list of users in a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    Returns:
      list: list of User objects.

    """
    users = []
    user_client = galaxy.users.UserClient(gi)
    for user_data in user_client.get_users():
        users.append(User(user_data))
    return users

def list_users(gi,name=None,long_listing_format=False):
    """
    List users in Galaxy instance

    Arguments:
      gi    : Galaxy instance
      name  : optionally, only list matching emails/usernames
      long_listing_format (boolean): if True then use a
        long listing format when reporting items

    """
    users = get_users(gi)
    if name:
        name = name.lower()
        users = filter(lambda u: fnmatch.fnmatch(u.username.lower(),name) or
                       fnmatch.fnmatch(u.email.lower(),name),users)
    users.sort(key=lambda u: u.email)
    for user in users:
        if long_listing_format:
            user = User(galaxy.users.UserClient(gi).show_user(user.id))
            print '\t'.join([str(x) for x in (user.email,
                                              user.username,
                                              user.nice_total_disk_usage,
                                              ("%s%%" % user.quota_percent
                                               if user.quota_percent
                                               else "0%"),
                                              ('admin' if user.is_admin
                                               else ''),
                                              user.id,)])
        else:
            print "%s\t%s" % (user.email,
                              user.username)
    print "total %s" % len(users)

def create_user(gi,email,username=None,passwd=None,only_check=False,
                mako_template=None):
    """
    Create a new Galaxy user

    Attempts to create a single user in a Galaxy instance with the
    supplied credentials.

    Arguments:
      gi    : Galaxy instance
      email : email address for the new user
      username: (optional) name to associate with the user. If
        'None' then will be generated from the email address.
      passwd: (optional) password for the new user. If 'None' then
        the user will be prompted to supply a password.
      only_check: if True then only run the checks, don't try to
        make the user on the system.
      mako_template (optional): Mako template that will be populated
        and printed

    Returns:
      0 on success, 1 on failure.

    """
    # Check if user already exists
    if not check_new_user_info(gi,email,username):
        return 1
    if only_check:
        print "Email and username ok: not currently in use"
        return 0
    # Prompt for password
    if passwd is None:
        try:
            passwd = get_passwd()
        except Exception, ex:
            logger.error("%s" % ex)
            return 1
    # Create the new user
    try:
        galaxy.users.UserClient(gi).create_local_user(username,
                                                      email,passwd)
    except galaxy.client.ConnectionError,ex:
        print "Failed to create user:"
        print ex
        return 1
    print "Created new account for %s" % email
    if mako_template:
        print render_mako_template(mako_template,email,passwd)
    return 0

def create_users_from_template(gi,template,start,end,passwd=None,
                               only_check=False):
    """
    Create a batch of users in Galaxy, based on template email

    Attempts to create multiple users in a Galaxy instance, using
    a template email address and a range of integer indices to
    generate the names.

    'template' should include a '#' symbol indicating where an integer
    index should be substituted (e.g. 'student#@galaxy.ac.uk').
    'start' and 'end' are the range of ids to create (e.g. 1 to 10).

    All accounts will be created with the same password; names will
    be generated automatically from the email addresses

    For example: the template 'student#@galaxy.ac.uk' with a range of
    1 to 5 will generate:

    student1@galaxy.ac.uk
    student2@galaxy.ac.uk
    ...
    student5@galaxy.ac.uk

    Arguments:
      gi       : Galaxy instance
      template : template email address for the batch of new users
      start    : initial integer index for user names
      end      : final integer index for user names
      passwd   : (optional) password for the new user. If 'None' then
        the user will be prompted to supply a password.
      only_check: if True then only run the checks, don't try to
        make the users on the system.

    Returns:
      0 on success, 1 on failure.
    
    """
    # Check template
    name,domain = template.split('@')
    if name.count('#') != 1 or domain.count('#') != 0:
        logger.error("Incorrect email template format")
        return 1
    # Deal with password
    if passwd is not None:
        if not validate_password(passwd):
            logger.error("Invalid password")
            return 1
    else:
        try:
            passwd = get_passwd()
        except Exception, ex:
            logger.error("%s" % ex)
            return 1
    # Generate emails
    emails = [template.replace('#',str(i)) for i in range(start,end+1)]
    # Check that these are available
    print "Checking availability"
    for email in emails:
        name = get_username_from_login(email)
        ##print "%s, %s" % (email,name)
        if not check_new_user_info(gi,email,name):
            return 1
    if only_check:
        print "All emails and usernames ok: not currently in use"
        return 0
    # Make the accounts
    for email in emails:
        name = get_username_from_login(email)
        print "Email : %s" % email
        print "Name  : %s" % name
        if create_user(gi,email,name,passwd):
            return 1
    return 0

def create_batch_of_users(gi,tsv,only_check=False,mako_template=None):
    """
    Create a batch of users in Galaxy from a list in a TSV file

    Attempts to create multiple users in a Galaxy instance, using
    a list of email addresses, passwords and (optionally) names
    supplied via a TSV file.

    The file should consist of lines of the form e.g.:

    a.user@galaxy.ac.uk	p@ssw0rd	a-user

    The last value (the public name) can be missing, in which case
    the name will be generated from the email address.

    If an email address is already used for an account in the
    target Galaxy instance then it will be skipped.

    Blank lines and lines starting with '#' are ignored.

    Arguments:
      gi : Galaxy instance
      tsv: Name of TSV file to read user data from
      only_check: if True then only run the checks, don't try to
        make the users on the system.
      mako_template (optional): Mako template that will be populated
        and printed

    Returns:
      0 on success, 1 on failure.
    
    """
    # Open file
    print "Reading data from file '%s'" % tsv
    users = {}
    for line in open(tsv,'r'):
        # Skip blank or comment lines
        if line.startswith('#') or not line.strip():
            continue
        # Extract data
        items = line.strip().split('\t')
        passwd = None
        name = None
        try:
            email = items[0].lower().strip()
            passwd = items[1].strip()
            name = items[2].strip()
        except IndexError:
            pass
        # Do checks
        if email in users:
            logger.error("%s: appears multiple times" % email)
            return 1
        if passwd is None:
            logger.error("%s: no password supplied" % email)
            return 1
        elif not validate_password(passwd):
            logger.error("%s: invalid password\n" % email)
            return 1
        if name is None:
            name = get_username_from_login(email)
        if check_new_user_info(gi,email,name):
            users[email] = { 'name': name, 'passwd': passwd }
            print "%s\t%s\t%s" % (email,'*****',name)
    if only_check:
        return 0
    # Make the accounts
    for email in users:
        name = users[email]['name']
        passwd = users[email]['passwd']
        if create_user(gi,email,name,passwd):
            return 1
        if mako_template:
            print render_mako_template(mako_template,email,passwd)
    return 0

def check_new_user_info(gi,email,username):
    """
    Check if username or login are already in use

    """
    lookup_user = filter(lambda u: u.email == email or
                         u.username == username,get_users(gi))
    if lookup_user:
        error_msg = "User details clash with existing user(s):"
        for user in lookup_user:
            error_msg += "\n%s" % ('\t'.join([user.email,
                                             user.username]))
        logger.error(error_msg)
        return False
    return True

def get_user_api_key(gi,username=None):
    """
    Retrieve an API key for a user

    Arguments:
      username (str): email address or username for the
        user to get the API key for; default is to get
        the key for the current user

    """
    if username is None:
        # Fetch the details for the current user
        try:
            user = galaxy.users.UserClient(gi).get_current_user()
            print "Username: %s" % username
        except galaxy.client.ConnectionError:
            logger.error("Cannot determine user associated with "
                          "this instance")
            return
        except AttributeError:
            logger.error("Unable to fetch user associated with "
                         "this instance")
            return
    else:
        # Fetch the details for the specified user
        user = None
        for u in get_users(gi):
            if (u.email == username) or (u.username == username):
                user = u
                break
    if user is None:
        logger.error("Cannot get info for user '%s'\n" % username)
        return
    # Get the API key
    user_id = user.id
    try:
        api_key = galaxy.users.UserClient(gi).create_user_apikey(user_id)
    except galaxy.client.ConnectionError,ex:
        print "Failed to fetch API key for user '%s': " % username
        print ex
        return
    return api_key

def check_username_format(username):
    """
    Check that format of 'username' is valid

    """
    return bool(re.match(r"^[a-z0-9\-]+$",username))

def get_username_from_login(email):
    """
    Create a public user name from an email address

    """
    return str(email).split('@')[0].lower().replace('.','-').replace('_','-')

def validate_password(passwd):
    """
    Check if password format is valid

    """
    if len(passwd) < 6:
        return False
    return True

def get_passwd():
    """Prompt user for a password

    Prompts user to enter and confirm a password. Raises an exception
    if the password is deemed to be invalid (e.g. too short), or if
    the password confirmation fails.

    Returns:
      Password string entered by the user.

    """
    passwd = getpass.getpass("Enter password for new user: ")
    if not validate_password(passwd):
        raise Exception("Invalid password: must be 6 or more characters")
    passwd2 = getpass.getpass("Confirm password: ")
    if passwd2 != passwd:
        raise Exception("Passwords don't match")
    return passwd

def render_mako_template(filename,email,password=None):
    """Render Mako template

    Render a Mako template from a file. The following variables are
    supplied to the template:

    first_name
    email
    password

    """
    first_name = email.split('.')[0].title()
    return Template(filename=filename).render(first_name=first_name,
                                              email=email,
                                              password=password)



#!/usr/bin/env python
#
# deprecated_cli: deprecated CLI commands
import sys
import os
import optparse
import logging
from nebulizer import get_version
from .core import get_galaxy_instance
from .cli import handle_ssl_warnings
from .cli import handle_debug
from .cli import handle_suppress_warnings
from .cli import handle_credentials
import users
import libraries
import tools

logger = logging.getLogger(__name__)

def deprecation_warning():
    """
    Issue a deprecation warning
    """
    logger.warning("*********** THIS UTILITY IS NOW DEPRECATED ***********")
    logger.warning("*********** Use 'nebulizer CMD...' instead ***********")

def base_parser(usage=None,description=None):
    """
    Create base parser with common options

    """
    p = optparse.OptionParser(usage=usage,
                              version="%s" % get_version(),
                              description=description)
    p.add_option('-k','--api_key',action='store',dest='api_key',
                 default=None,
                 help="specify API key to use for connecting to "
                 "Galaxy instance. Must be supplied if there is "
                 "no API key stored for the specified instance, "
                 "(unless --username option is specified). If "
                 "there is a stored API key this overrides it.")
    p.add_option('-u','--username',action='store',dest='username',
                 default=None,
                 help="specify username (i.e. email) for connecting "
                 "to Galaxy instance, as an alternative to using "
                 "the API key. Prompts for a password unless one "
                 "is supplied via the --galaxy_password option.")
    p.add_option('-P','--galaxy_password',action='store',
                 dest='galaxy_password',default=None,
                 help="supply password for connecting to Galaxy "
                 "instance, when using the --username option.")
    p.add_option('-n','--no-verify',action='store_true',dest='no_verify',
                 default=False,help="don't verify HTTPS "
                 "connections. Use this when connecting to a Galaxy "
                 "instance which uses self-signed certificates.")
    p.add_option('-q','--suppress-warnings',action='store_true',
                 dest='suppress_warnings',
                 default=False,help="suppress warning messages "
                 "from nebulizer")
    p.add_option('--debug',action='store_true',dest='debug',
                 default=False,help="turn on debugging output")
    return p
    
def manage_users(args=None):
    """
    Implements the 'manage_users' utility

    """
    deprecation_warning()
    if args is None:
        args = sys.argv[1:]

    p = base_parser(usage=\
                    "\n\t%prog list   GALAXY_URL [options]"
                    "\n\t%prog create GALAXY_URL EMAIL [PUBLIC_NAME]"
                    "\n\t%prog create GALAXY_URL -t TEMPLATE START [END]"
                    "\n\t%prog create GALAXY_URL -b FILE [options]",
                    description="Manage and create users in a Galaxy "
                    "instance")
    commands = ['list','create']

    # Get compulsory arguments
    if len(args) == 1:
        if args[0] == '-h' or args[0] == '--help':
            p.print_usage()
        elif args[0] == '--version':
            p.print_version()
        sys.exit(0)
    if len(args) < 2:
        p.error("need to supply a command and a Galaxy URL/alias")
    command = args[0]
    galaxy_url = args[1]

    # Setup additional command line options
    if command not in commands:
        p.error("unrecognised command: '%s'" % command)
    elif command == 'list':
        p.set_usage("%prog list GALAXY_URL [options]")
        p.add_option('--name',action='store',dest='name',default=None,
                     help="specific emails/user name(s) to list")
        p.add_option('-l',action='store_true',
                     dest='long_listing_format',default=False,
                     help="use a long listing format (include ids, "
                     "disk usage and admin status)")
    elif command == 'create':
        p.set_usage("\n\t%prog create GALAXY_URL EMAIL [PUBLIC_NAME]"
                    "\n\t%prog create GALAXY_URL -t TEMPLATE START [END]"
                    "\n\t%prog create GALAXY_URL -b FILE [options]")
        p.add_option('-p','--password',action='store',dest='passwd',
                     default=None,
                     help="specify password for new user account "
                     "(otherwise program will prompt for password)")
        p.add_option('-c','--check',action='store_true',dest='check',
                     default=False,
                     help="check user details but don't try to create "
                     "the new account")
        p.add_option('-t','--template',action='store_true',
                     dest='template',default=False,
                     help="indicates that EMAIL is actually a "
                     "'template' email address which includes a '#' "
                     "symbol as a placeholder where an integer index "
                     "should be substituted to make multiple accounts "
                     "(e.g. 'student#@galaxy.ac.uk').")
        p.add_option('-b','--batch',action='store_true',dest='batch',
                     default=False,
                     help="create multiple users reading details from "
                     "TSV file (columns should be: "
                     "email,password[,public_name])")
        p.add_option('-m','--message',action='store',
                     dest='message_template',default=None,
                     help="populate and output Mako template "
                     "MESSAGE_TEMPLATE")
        
    # Process remaining arguments on command line
    if args[1] in ('-h','--help','--version'):
        args = args[1:]
    else:
        args = args[2:]
    options,args = p.parse_args(args)
    handle_debug(debug=options.debug)
    handle_suppress_warnings(suppress_warnings=options.suppress_warnings)
    handle_ssl_warnings(verify=(not options.no_verify))

    # Handle password if required
    email,password = handle_credentials(options.username,
                                        options.galaxy_password,
                                        prompt="Password for %s: " % galaxy_url)

    # Get a Galaxy instance
    gi = get_galaxy_instance(galaxy_url,api_key=options.api_key,
                             email=email,password=password,
                             verify_ssl=(not options.no_verify))
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        sys.exit(1)

    # Execute command
    if command == 'list':
        users.list_users(gi,name=options.name,long_listing_format=
                         options.long_listing_format)
    elif command == 'create':
        # Check message template is .mako file
        if options.message_template:
            if not os.path.isfile(options.message_template):
                logger.critical("Message template '%s' not found"
                                % options.message_template)
                sys.exit(1)
            elif not options.message_template.endswith(".mako"):
                logger.critical("Message template '%s' is not a .mako file"
                                % options.message_template)
                sys.exit(1)
        if options.template:
            # Get the template and range of indices
            template = args[0]
            start = int(args[1])
            try:
                end = int(args[2])
            except IndexError:
                end = start
            # Create users
            retval = users.create_users_from_template(gi,template,
                                                      start,end,options.passwd,
                                                      only_check=options.check)
        elif options.batch:
            # Get the file with the user data
            tsvfile = args[0]
            # Create users
            retval = users.create_batch_of_users(gi,tsvfile,
                                                 only_check=options.check,
                                                 mako_template=options.message_template)
        else:
            # Collect email and (optionally) public name
            email = args[0]
            try:
                name = args[1]
                if not users.check_username_format(name):
                    logger.critical("Invalid name: must contain only "
                                    "lower-case letters, numbers and "
                                    "'-'")
                    sys.exit(1)
            except IndexError:
                # No public name supplied, make from email address
                name = users.get_username_from_login(email)
            # Create user
            print "Email : %s" % email
            print "Name  : %s" % name
            retval = users.create_user(gi,email,name,options.passwd,
                                       only_check=options.check,
                                       mako_template=options.message_template)

def manage_libraries(args=None):
    """
    Implements the 'manage_libraries' utility

    """
    deprecation_warning()
    if args is None:
        args = sys.argv[1:]

    p = base_parser(usage=\
                    "\n\t%prog list GALAXY_URL [options]"
                    "\n\t%prog create_library GALAXY_URL [options] NAME"
                    "\n\t%prog create_folder GALAXY_URL [options] PATH"
                    "\n\t%prog add_datasets GALAXY_URL [options] DEST FILE...",
                    description="Manage and populate data libraries in a "
                    "Galaxy instance")
    
    commands = ['list','create_library','create_folder','add_datasets']

    # Get compulsory arguments
    if len(args) == 1:
        if args[0] == '-h' or args[0] == '--help':
            p.print_usage()
        elif args[0] == '--version':
            p.print_version()
        sys.exit(0)
    if len(args) < 2:
        p.error("need to supply a command and a Galaxy URL/alias")
    command = args[0]
    galaxy_url = args[1]

    # Setup additional command line options
    if command not in commands:
        p.error("unrecognised command: '%s'" % command)
    elif command == 'list':
        p.set_usage("%prog list GALAXY_URL [options]")
        p.add_option('-l',action='store_true',
                     dest='long_listing_format',default=False,
                     help="use a long listing format (include ids, "
                     "descriptions and file sizes and paths)")
    elif command == 'create_library':
        p.set_usage("%prog create_library GALAXY_URL [options] NAME")
        p.add_option('-d','--description',action='store',
                     dest='description',default=None,
                     help="optional description")
        p.add_option('-s','--synopsis',action='store',dest='synopsis',
                     default=None,help="optional synopsis")
    elif command == 'create_folder':
        p.set_usage("%prog create_folder GALAXY_URL [options] PATH")
        p.add_option('-d','--description',action='store',
                     dest='description',default=None,
                     help="optional description")
    elif command == 'add_datasets':
        p.set_usage("%prog add_datasets GALAXY_URL [options] DEST FILE...")
        p.add_option('--server',action='store_true',dest='from_server',
                     default=False,
                     help="upload files from Galaxy server file system "
                     "paths (default is to upload files from local "
                     "system)")
        p.add_option('--link',action='store_true',dest='link',
                     default=False,
                     help="create symlinks to files on server (only "
                     "valid if used with --server; default is to copy "
                     "files into Galaxy)")
        p.add_option('--dbkey',action='store',dest='dbkey',default='?',
                     help="dbkey to assign to files (default is '?')")
        p.add_option('--file_type',action='store',dest='file_type',
                     default='auto',
                     help="file type to assign to files (default is "
                     "'auto')")

    # Process remaining arguments on command line
    if args[1] in ('-h','--help','--version'):
        args = args[1:]
    else:
        args = args[2:]
    options,args = p.parse_args(args)
    handle_debug(debug=options.debug)
    handle_suppress_warnings(suppress_warnings=options.suppress_warnings)
    handle_ssl_warnings(verify=(not options.no_verify))

    # Handle password if required
    email,password = handle_credentials(options.username,
                                        options.galaxy_password,
                                        prompt="Password for %s: " % galaxy_url)

    # Get a Galaxy instance
    gi = get_galaxy_instance(galaxy_url,api_key=options.api_key,
                             email=email,password=password,
                             verify_ssl=(not options.no_verify))
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        sys.exit(1)

    # Execute command
    if command == 'list':
        if len(args) == 1:
            # List folders in data library
            libraries.list_library_contents(gi,args[0],long_listing_format=
                                            options.long_listing_format)
        else:
            # List data libraries
            libraries.list_data_libraries(gi)
    elif command == 'create_library':
        # Create a new data library
        if len(args) == 1:
            libraries.create_library(gi,args[0],
                                     description=options.description,
                                     synopsis=options.synopsis)
        else:
            p.error("Usage: create_library NAME")
    elif command == 'create_folder':
        # Create a new folder data library
        if len(args) == 1:
            libraries.create_folder(gi,args[0],
                                    description=options.description)
        else:
            p.error("Usage: create_folder PATH")
    elif command == 'add_datasets':
        # Add a dataset to a library
        if len(args) < 2:
            p.error("Usage: add_datasets DEST FILE [FILE...]")
        libraries.add_library_datasets(gi,args[0],args[1:],
                                       from_server=options.from_server,
                                       link_only=options.link,
                                       file_type=options.file_type,
                                       dbkey=options.dbkey)
 
def manage_tools(args=None):
    """
    Implements the 'manage_tools' utility

    """
    deprecation_warning()
    if args is None:
        args = sys.argv[1:]

    p = base_parser(usage=\
                    "\n\t%prog list GALAXY_URL [options]"
                    "\n\t%prog installed GALAXY_URL [options]"
                    "\n\t%prog tool_panel GALAXY_URL [options]"
                    "\n\t%prog install GALAXY_URL [options] SHED OWNER TOOL [REVISION]"
                    "\n\t%prog update GALAXY_URL [options] SHED OWNER TOOL",
                    description="Manage and install tools in a Galaxy "
                    "instance")
    
    commands = ['list','installed','tool_panel','install','update']

    # Get compulsory arguments
    if len(args) == 1:
        if args[0] == '-h' or args[0] == '--help':
            p.print_usage()
        elif args[0] == '--version':
            p.print_version()
        sys.exit(0)
    if len(args) < 2:
        p.error("need to supply a command and a Galaxy URL/alias")
    command = args[0]
    galaxy_url = args[1]

    # Setup additional command line options
    if command not in commands:
        p.error("unrecognised command: '%s'" % command)
    elif command == 'list':
        p.set_usage("%prog list GALAXY_URL [options]")
        p.add_option('--name',action='store',dest='name',default=None,
                     help="specific tool name(s) to list")
        p.add_option('--installed',action='store_true',
                     dest='installed_only',default=False,
                     help="only list tools installed from a toolshed")
    elif command == 'installed':
        p.set_usage("%prog installed GALAXY_URL [options]")
        p.add_option('--name',action='store',dest='name',default=None,
                     help="specific tool repository/ies to list")
        p.add_option('--toolshed',action='store',dest='toolshed',default=None,
                     help="only list repositories from TOOLSHED")
        p.add_option('--owner',action='store',dest='owner',default=None,
                     help="only list repositories owned by OWNER")
        p.add_option('--list-tools',action='store_true',dest='list_tools',
                     default=None,
                     help="list the associated tools for each repository")
        p.add_option('--updateable',action='store_true',dest='updateable',
                     default=None,
                     help="only show repositories with uninstalled updates "
                     "or upgrades")
    elif command == 'tool_panel':
        p.set_usage("%prog tool_panel GALAXY_URL [options]")
        p.add_option('--name',action='store',dest='name',default=None,
                     help="specific tool panel section(s) to list")
        p.add_option('--list-tools',action='store_true',dest='list_tools',
                     default=None,
                     help="list the associated tools for each section")
    elif command == 'install':
        p.set_usage("%prog install GALAXY_URL [options] SHED OWNER TOOL "
                    "[REVISION]")
        p.add_option('--tool-panel-section',action='store',
                     dest='tool_panel_section',default=None,
                     help="tool panel section name or id to install the "
                     "tool under")
    elif command == 'update':
        p.set_usage("%prog update GALAXY_URL [options] SHED OWNER TOOL")

    # Process remaining arguments on command line
    if args[1] in ('-h','--help','--version'):
        args = args[1:]
    else:
        args = args[2:]
    options,args = p.parse_args(args)
    handle_debug(debug=options.debug)
    handle_suppress_warnings(suppress_warnings=options.suppress_warnings)
    handle_ssl_warnings(verify=(not options.no_verify))

    # Handle password if required
    email,password = handle_credentials(options.username,
                                        options.galaxy_password,
                                        prompt="Password for %s: " % galaxy_url)

    # Get a Galaxy instance
    gi = get_galaxy_instance(galaxy_url,api_key=options.api_key,
                             email=email,password=password,
                             verify_ssl=(not options.no_verify))
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        sys.exit(1)

    # Execute command
    if command == 'list':
        tools.list_tools(gi,name=options.name,
                         installed_only=options.installed_only)
    elif command == 'installed':
        tools.list_installed_repositories(gi,name=options.name,
                                          toolshed=options.toolshed,
                                          owner=options.owner,
                                          list_tools=options.list_tools,
                                          only_updateable=options.updateable)
    elif command == 'tool_panel':
        tools.list_tool_panel(gi,name=options.name,
                              list_tools=options.list_tools)
    elif command == 'install':
        if len(args) < 3:
            p.error("Need to supply toolshed, owner, repo and optional "
                    "revision")
        toolshed,owner,repo = args[:3]
        if len(args) == 4:
            revision = args[3]
        else:
            revision = None
        status = tools.install_tool(
            gi,toolshed,repo,owner,revision=revision,
            tool_panel_section=options.tool_panel_section)
        sys.exit(status)
    elif command == 'update':
        if len(args) != 3:
            p.error("Need to supply toolshed, owner and repo")
        toolshed,owner,repo = args[:3]
        status = tools.update_tool(gi,toolshed,repo,owner)
        sys.exit(status)



#!/usr/bin/env python
#
# libraries: functions for managing data libraries
import logging
import os
import fnmatch
from .core import get_current_user
from bioblend import galaxy
import logging

logger = logging.getLogger(__name__)

def list_data_libraries(gi):
    """
    Return list of data libraries

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    """
    for lib in galaxy.libraries.LibraryClient(gi).get_libraries():
        print "%s\t%s\t%s" % (lib['name'],lib['description'],lib['id'])
        ##print "%s" % lib

def library_id_from_name(gi,library_name):
    """
    Fetch ID for data library from library name

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      library_name (str): name of data library to look up

    Returns:
      str: ID for data library, or None if name not found.

    """
    try:
        return galaxy.libraries.LibraryClient(gi).get_libraries(
            name=library_name)[0]['id']
    except IndexError:
        return None

def folder_id_from_name(gi,library_id,folder_name):
    """
    Fetch ID for folder

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      library_id (str): ID for parent data library
      folder_name (str): name of folder to look up

    Returns:
      str: ID for folder, or None if name not found.

    """
    lib_client = galaxy.libraries.LibraryClient(gi)
    folder_name = normalise_folder_path(folder_name)
    logger.debug("Looking for '%s' in library %s" % (folder_name,
                                                      library_id))
    for folder in lib_client.get_folders(library_id):
        logger.debug("Checking '%s'" % folder['name'])
        if folder['name'] == folder_name:
            return folder['id']
    return None

def list_library_contents(gi,path,long_listing_format=False):
    """
    Print contents of folder in data library

    The output of this function is loosely based on the
    output from the Linux `ls` command, and depends on
    what data items the input path (which can include
    wildcards).

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      path (str): path describing a data library or
        a folder in a library
      long_listing_format (boolean): if True then use a
        long listing format when reporting items

    """
    # Get name and id for parent data library
    logger.debug("Path '%s'" % path)
    lib_client = galaxy.libraries.LibraryClient(gi)
    library_name,folder_path = split_library_path(path)
    logger.debug("library_name '%s'" % library_name)
    library_id = library_id_from_name(gi,library_name)
    if library_id is None:
        print "No library '%s'" % library_name
        return
    # Get library contents
    library_contents = lib_client.show_library(library_id,contents=True)
    logger.debug("folder_path '%s'" % folder_path)
    # Bioblend class for getting more info for datasets if
    # using a long listing format
    dataset_client = galaxy.datasets.DatasetClient(gi)
    # Determine if we're matching against a wildcard pattern
    pattern = folder_path
    wildcard_pattern = False
    for c in "*?[]":
        try:
            pattern.index(c)
            wildcard_pattern = True
        except ValueError:
            pass
    # Output mode depends on whether we have wildcards
    if not wildcard_pattern:
        # Exact matches only
        matches = filter(lambda x: x['name'] == pattern,
                         library_contents)
        if not matches:
            logger.error("Cannot access %s: no matching libraries "
                         "or folders" % path)
            return
        for item in matches:
            if item['type'] == 'folder':
                contents = filter(lambda x: os.path.split(x['name'])[0]
                                  == item['name'],library_contents)
            else:
                contents = matches
        # Remove 'root' folder
        contents = filter(lambda x: x['name'] != '/',contents)
        # Report
        if long_listing_format:
            print "total %s" % len(contents)
        for item in contents:
            if item['type'] == 'folder':
                report_folder(lib_client.show_folder(library_id,
                                                     item['id']),
                                    long_listing=long_listing_format)
            else:
                report_dataset(dataset_client.show_dataset(item['id'],
                                                           hda_ldda='ldda'),
                               long_listing=long_listing_format)
    else:
        # Number of levels to match
        nlevels = pattern.count('/')
        # Mixture of matches possible
        matches = filter(lambda x: fnmatch.fnmatch(x['name'],pattern)
                         and x['name'].count('/') == nlevels,
                         library_contents)
        if not matches:
            logger.error("Cannot access %s: no matching libraries "
                         "or folders\n" % path)
            return
        # Identify the folders that are matched exactly
        folders = []
        for item in matches:
            if item['type'] == 'folder' and item['name'] != '/':
                folders.append(item)
        # Locate non-folder items that are not in any of
        # the folders previously identified
        datasets = []
        for item in matches:
            implicit_dataset = False
            if item['type'] == 'folder':
                continue
            else:
                parent = os.path.split(item['name'])[0]+'/'
                for folder in folders:
                    if folder['name'].startswith(parent):
                        print "Parent = %s" % parent
                        implicit_dataset = True
                        break
            # Item outside of folders
            if not implicit_dataset:
                datasets.append(item)
        # List the datasets
        for dataset in datasets:
            report_dataset(dataset_client.show_dataset(item['id'],
                                                       hda_ldda='ldda'),
                           long_listing=long_listing_format)
        # List the contents of each folder
        for folder in folders:
            print "\n%s:" % folder['name']
            folder_contents = filter(lambda x: os.path.split(x['name'])[0]
                                     == folder['name'],library_contents)
            if long_listing_format:
                print "total %s" % len(folder_contents)
            for item in folder_contents:
                if item['type'] == 'folder':
                    report_folder(lib_client.show_folder(library_id,
                                                         item['id']),
                    long_listing=long_listing_format)
                else:
                    report_dataset(dataset_client.show_dataset(item['id'],
                                                               hda_ldda='ldda'),
                                   long_listing=long_listing_format)

def create_library(gi,name,description=None,synopsis=None):
    """
    Create a new data library

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      name (str): name of the new data library
      description (str): (optional) text to add as the
        description of the new data library
      synopsis (str): (optional) short 'synopsis' text
        to associate with the library

    Returns:
      str: id for data library.

    """
    lib_client = galaxy.libraries.LibraryClient(gi)
    if library_id_from_name(gi,name):
        print "Target data library already exists"
        return library_id_from_name(gi,name)
    library = lib_client.create_library(name,
                                        description=description,
                                        synopsis=synopsis)
    #print "%s" % library
    return library['id']

def create_folder(gi,path,description=None):
    """
    Create a new folder in a data library

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      path (str): path to the new folder (must be within
        an existing data library)
      description (str): text to add as the description
        of the new folder

    Returns:
      str: id for new folder.

    """
    # Break up the path
    library_name,folder_path = split_library_path(path)
    logger.debug("library_name: %s" % library_name)
    logger.debug("folder_path : %s" % folder_path)
    # Get name and id for parent data library
    lib_client = galaxy.libraries.LibraryClient(gi)
    library_id = library_id_from_name(gi,library_name)
    if library_id is None:
        print "Top level data library '%s' not found" % library_name
        return None
    # Get folder name and base folder
    folder_base,folder_name = os.path.split(folder_path)
    # Check folder with same name doesn't already exist
    folder_id = folder_id_from_name(gi,library_id,folder_path)
    #print "folder_id '%s' for '%s'" % (folder_id,folder_path)
    if folder_id_from_name(gi,library_id,folder_path):
        print "Target folder already exists"
        return None
    #print "folder_name '%s' folder_base '%s'" % (folder_name,
    #                                             folder_base)
    base_folder_id = folder_id_from_name(gi,library_id,folder_base)
    #print "base_folder_id %s" % base_folder_id
    new_folder = lib_client.create_folder(library_id,folder_name,
                                          description=description,
                                          base_folder_id=base_folder_id)
    #print "%s" % new_folder
    return new_folder[0]['id']

def add_library_datasets(gi,path,files,
                         from_server=False,
                         link_only=False,
                         file_type='auto',
                         dbkey='?'):
    """
    Add datasets to a data library

    Note that the Galaxy instance must be associated with a
    real user in order to upload data; 'userless' instances
    (for example if using the master API) will be refused.

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      path (str): path of folder to add datasets to
        files (list): list of files to upload to the data
        library
      from_server (bool): set True if files to be added are
        on the Galaxy server filesystem (default is False
        i.e. files on the local file system)
      link_only (bool): for files on the Galaxy server
        filesystem, set True to create links to the files
        (default is False i.e. copy files into Galaxy)
      file_type (str): explicit Galaxy file type to apply
        to all uploaded files (default is 'auto')
      dbkey (str): explicit dbkey to apply to all uploaded
        files (default is '?')

    """
    # Check that we're not using a 'userless' API key (e.g.
    # master key)
    if get_current_user(gi) is None:
        logger.error("No user associated with this API key, "
                     "data upload aborted")
        return
    # Break up the path
    library_name,folder_path = split_library_path(path)
    # Get name and id for parent data library
    lib_client = galaxy.libraries.LibraryClient(gi)
    library_id = library_id_from_name(gi,library_name)
    print "Library name '%s' id '%s'" % (library_name,library_id)
    # Get id for parent folder
    folder_id = folder_id_from_name(gi,library_id,folder_path)
    print "Folder name '%s' id '%s'" % (folder_path,folder_id)
    if from_server:
        # Assume that files are on Galaxy fileserver not localhost
        filesystem_paths = '\n'.join(files)
        print "Uploading files from Galaxy server:"
        print "%s" % filesystem_paths
        lib_client.upload_from_galaxy_filesystem(
            library_id,filesystem_paths,
            folder_id=folder_id,
            file_type=file_type,
            dbkey=dbkey,
            link_data_only=link_only,
            roles='')
    else:
        # Files are on localhost
        for f in files:
            print "Uploading file '%s'" % f
            lib_client.upload_file_from_local_path(
                library_id,f,
                folder_id=folder_id,
                file_type=file_type,
                dbkey=dbkey)

def split_library_path(path):
    """
    Split library path into library and folder components

    Note that the folder path will be returned as the
    normalised path i.e. /path/to/folder (single leading
    slash with no trailing slash).

    Arguments:
      path (str): path describing a folder in a data library

    Returns:
      Tuple: (library_name,folder_name)

    """
    components = path.strip('/').split('/')
    library_name = components[0]
    if len(components) == 1:
        folder_name = ''
    else:
        folder_name = '/'.join(components[1:])
    return (library_name,
            normalise_folder_path(folder_name))

def normalise_folder_path(path):
    """
    Normalise a folder path

    Normalise folder paths so that they are always of the
    form

    /path/to/folder

    i.e a single leading slash with no trailing slash

    Arguments:
      path (str): path for a data library folder

    Returns:
      str: normalised folder path.

    """
    return '/'+'/'.join(filter(lambda x: x != '',path.split('/')))

def report_folder(folder_data,long_listing=False):
    """
    Report details of a library folder

    Arguments:
      folder_data (dict): dictionary returned from an
        appropriate call to bioblend
      long_listing_format (boolean): if True then use a
        long listing format when reporting items

    """
    logger.debug("%s" % folder_data)
    if long_listing:
        print "%s/\t%s\t%s" % (folder_data['name'],
                                folder_data['description'],
                                folder_data['id'])
    else:
        print "%s/" % os.path.split(folder_data['name'])[1]

def report_dataset(dataset_data,long_listing=False):
    """
    Report details of a library dataset

    Arguments:
      dataset_data (dict): dictionary returned from an
        appropriate call to bioblend
      long_listing_format (boolean): if True then use a
        long listing format when reporting items

    """
    logger.debug("%s" % dataset_data)
    if long_listing:
        print '\t'.join([str(x) for x in (dataset_data['name'],
                                          dataset_data['file_ext'],
                                          dataset_data['genome_build'],
                                          dataset_data['file_size'],
                                          dataset_data['file_name'],)])
    else:
        print "%s" % os.path.split(dataset_data['name'])[1]



#!/usr/bin/env python
#
# cli: functions for building command utilities
import getpass
import logging
import click
import time
from nebulizer import get_version
from .core import get_galaxy_instance
from .core import get_current_user
from .core import ping_galaxy_instance
from .core import turn_off_urllib3_warnings
from .core import Credentials
import users
import libraries
import tools

# Initialise logging
logger = logging.getLogger(__name__)
# Suppress errors from bioblend
logging.getLogger("bioblend").setLevel(logging.CRITICAL)

def handle_ssl_warnings(verify=True):
    """
    Turn off SSL warnings from urllib3

    Arguments:
      verify (bool): if False then disable the warnings from
        urllib3 about SSL certificate verification

    """
    if not verify:
        logger.warning("SSL certificate verification has "
                       "been disabled")
        turn_off_urllib3_warnings()

def handle_debug(debug=True):
    """
    Turn on debugging output from logging

    Arguments:
      debug (bool): if True then turn on debugging output

    """
    if debug:
        level = logging.DEBUG
    else:
        level = logging.WARNING
    logging.getLogger("nebulizer").setLevel(level)

def handle_suppress_warnings(suppress_warnings=True):
    """
    Suppress warning messages output from logging

    Arguments:
      suppress_warnings (bool): if True then turn off
        warning messages

    """
    if suppress_warnings:
        logging.getLogger("nebulizer").setLevel(logging.ERROR)

def handle_credentials(email,password,prompt="Password: "):
    """
    Sort out email and password for accessing Galaxy

    Arguments:
      email (str): Galaxy e-mail address corresponding to the user
      password (str): password of Galaxy account corresponding to
        email address; if None then user will be prompted to
        supply password on the command line
      prompt (str): text to display as password prompt

    Returns:
      Tuple: tuple consisting of (email,password).

    """
    if email is None:
        return (None,None)
    if password is None:
        password = getpass.getpass(prompt)
    return (email,password)

def fetch_new_api_key(galaxy_url,email,password=None,verify=True):
    """
    Fetch a new API key from a Galaxy instance

    Arguments:
      galaxy_url (str): alias or URL of Galaxy instance to get
        API key for
      email (str): Galaxy e-mail address corresponding to the
        user to fetch API for
      password (str): password of Galaxy account corresponding to
        email address (optional)
      verify (boolean): if False then disable SSL verification
        when connecting to Galaxy instance (default is to keep
        SSL verification)

    """
    print "Fetching API key from %s" % galaxy_url
    email,password = handle_credentials(
        email,password,
        prompt="Please supply password for %s: " % galaxy_url)
    gi = get_galaxy_instance(galaxy_url,
                             email=email,password=password,
                             verify_ssl=verify)
    return users.get_user_api_key(gi,username=email)

class Context(object):
    """
    Provide context for nebulizer command
    """
    def __init__(self):
        self.api_key = None
        self.username = None
        self.galaxy_password = None
        self.no_verify = False
        self.debug = False

    def galaxy_instance(self,alias):
        """
        Return Galaxy instance based on context

        Attempts to create a Bioblend based on the supplied
        arguments to the nebulizer command.
        """
        email,password = handle_credentials(
            self.username,
            self.galaxy_password,
            prompt="Password for %s: " % alias)
        gi = get_galaxy_instance(alias,api_key=self.api_key,
                                 email=email,password=password,
                                 verify_ssl=(not self.no_verify))
        return gi

pass_context = click.make_pass_decorator(Context,ensure=True)

@click.group()
@click.version_option(version=get_version())
@click.option('--api_key','-k',
              help="specify API key to use for connecting to "
              "Galaxy instance. Must be supplied if there is "
              "no API key stored for the specified instance, "
              "(unless --username option is specified). If "
              "there is a stored API key this overrides it.")
@click.option('--username','-u',
              help="specify username (i.e. email) for connecting "
              "to Galaxy instance, as an alternative to using "
              "the API key. Prompts for a password unless one "
              "is supplied via the --galaxy_password option.")
@click.option('--galaxy_password','-P',
              help="supply password for connecting to Galaxy "
              "instance, when using the --username option.")
@click.option('--no-verify','-n',is_flag=True,
              help="don't verify HTTPS connections when "
              "connecting to Galaxy instance. Use this when "
              "interacting with Galaxy instances which use "
              "self-signed certificates.")
@click.option('--suppress-warnings','-q',is_flag=True,
              help="suppress warning messages from nebulizer.")
@click.option('--debug',is_flag=True,
              help="turn on debugging output.")
@pass_context
def nebulizer(context,api_key,username,galaxy_password,
              no_verify,suppress_warnings,debug):
    """
    Manage users, tools and data libraries in Galaxy instances
    via the API
    """
    context.api_key = api_key
    context.username = username
    context.galaxy_password = galaxy_password
    context.no_verify = no_verify
    context.debug = debug
    context.suppress_warnings = suppress_warnings
    handle_debug(debug=context.debug)
    handle_suppress_warnings(suppress_warnings=context.suppress_warnings)
    handle_ssl_warnings(verify=(not context.no_verify))

@nebulizer.command()
@pass_context
def list_keys(context):
    """
    List stored Galaxy API keys.

    Prints a list of stored aliases with the associated
    Galaxy URLs and API keys.
    """
    instances = Credentials()
    for alias in instances.list_keys():
        galaxy_url,api_key = instances.fetch_key(alias)
        click.echo("%s\t%s\t%s" % (alias,galaxy_url,api_key))

@nebulizer.command()
@click.argument("alias")
@click.argument("galaxy_url")
@click.argument("api_key",required=False)
@pass_context
def add_key(context,alias,galaxy_url,api_key=None):
    """
    Store new Galaxy URL and API key.

    ALIAS is the name that the instance will be stored
    against; GALAXY_URL is the URL for the instance;
    API_KEY is the corresponding API key.

    If API_KEY is not supplied then nebulizer will
    attempt to fetch one automatically.
    """
    instances = Credentials()
    if alias in instances.list_keys():
        logger.error("'%s' already exists" % alias)
        return 1
    if api_key is None:
        # No API key supplied as argument, try to connect
        # to Galaxy and fetch directly
        gi = context.galaxy_instance(galaxy_url)
        if gi is None:
            logger.critical("%s: failed to connect" % galaxy_url)
            return 1
        api_key = gi.key
    # Store the entry
    instances.store_key(alias,galaxy_url,api_key)

@nebulizer.command()
@click.option('--new-url',
              help="specify new URL for Galaxy instance")
@click.option('--new-api-key',
              help="specify new API key for Galaxy instance")
@click.option('--fetch-api-key',is_flag=True,
              help="fetch new API key for Galaxy instance")
@click.argument("alias")
@pass_context
def update_key(context,alias,new_url,new_api_key,fetch_api_key):
    """
    Update stored Galaxy API key.

    Update the Galaxy URL and/or API key stored
    against ALIAS.
    """
    instances = Credentials()
    if alias not in instances.list_keys():
        logger.error("'%s': not found" % alias)
        return 1
    if new_url:
        galaxy_url = new_url
    else:
        galaxy_url = instances.fetch_key(alias)[0]
    click.echo("galaxy_url: %s" % galaxy_url)
    click.echo("username  : %s" % context.username)
    if fetch_api_key:
        # Attempt to connect to Galaxy and fetch API key
        gi = context.galaxy_instance(alias)
        if gi is None:
            logger.critical("%s: failed to connect" % alias)
            return 1
        new_api_key = gi.key
    instances.update_key(alias,
                         new_url=new_url,
                         new_api_key=new_api_key)

@nebulizer.command()
@click.argument("alias")
@pass_context
def remove_key(context,alias):
    """
    Remove stored Galaxy API key.

    Removes the Galaxy URL/API key pair associated with
    ALIAS from the list of stored keys.
    """
    instances = Credentials()
    instances.remove_key(alias)

@nebulizer.command()
@click.option("--name",
              help="list only users with matching email or user "
              "name. Can include glob-style wild-cards.")
@click.option("--long","-l","long_listing",is_flag=True,
              help="use a long listing format that includes ids,"
              " disk usage and admin status.")
@click.argument("galaxy")
@pass_context
def list_users(context,galaxy,name,long_listing):
    """
    List users in Galaxy instance.

    Prints details of user accounts in GALAXY instance.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # List users
    users.list_users(gi,name=name,long_listing_format=long_listing)

@nebulizer.command()
@click.option('--password','-p',
              help="specify password for new user account "
              "(otherwise program will prompt for password)")
@click.option('--check','-c','only_check',is_flag=True,
              help="check user details but don't try to create the "
              "new account")
@click.option('--message','-m','message_template',
              type=click.Path(exists=True),
              help="Mako template to populate and output")
@click.argument("galaxy")
@click.argument("email")
@click.argument("public_name",required=False)
@pass_context
def create_user(context,galaxy,email,public_name,password,only_check,
                message_template):
    """
    Create new Galaxy user.

    Creates a new user in GALAXY, using EMAIL for the username.
    If PUBLIC_NAME is not supplied then it will be generated
    automatically from the leading part of the email address.

    If a password for the new account is not supplied using the
    --password option then nebulizer will prompt for one.
    """
    # Check message template is a .mako file
    if message_template:
        if not message_template.endswith(".mako"):
            logger.critical("Message template '%s' is not a .mako file"
                            % message_template)
            return 1
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Sort out email and public name
    if public_name:
        if not users.check_username_format(public_name):
            logger.critical("Invalid public name: must contain only "
                            "lower-case letters, numbers and '-'")
            return 1
    else:
        # No public name supplied, make from email address
        public_name = users.get_username_from_login(email)
    # Create user
    print "Email : %s" % email
    print "Name  : %s" % public_name
    return users.create_user(gi,email,public_name,password,
                             only_check=only_check,
                             mako_template=message_template)

@nebulizer.command()
@click.option('--password','-p',
              help="specify password for new user accounts "
              "(otherwise program will prompt for password). "
              "All accounts will be created with the same "
              "password.")
@click.option('--check','-c','only_check',is_flag=True,
              help="check user details but don't try to create "
              "the new account.")
@click.argument("galaxy")
@click.argument("template")
@click.argument("start",type=int)
@click.argument("end",type=int,required=False)
@pass_context
def create_batch_users(context,galaxy,template,start,end,
                       password,only_check):
    """
    Create multiple Galaxy users from a template.

    Creates a batch of users in GALAXY using TEMPLATE; this
    should be a template email address which includes a
    '#' symbol as a placeholder for an integer index.

    The range of integers is defined by START and END; if
    only one of these is supplied then START is assumed to be
    1 and the supplied value is END.

    For example: using the template 'user#@example.org'
    with the range 1 to 5 will create new accounts:

    user1@galaxy.org
    ...
    user5@galaxy.org
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Sort out start and end indices
    if end is None:
        end = start
        start = 1
    # Create users
    return users.create_users_from_template(gi,template,
                                            start,end,password,
                                            only_check=only_check)

@nebulizer.command()
@click.option('--check','-c','only_check',is_flag=True,
              help="check user details but don't try to create the "
              "new account.")
@click.option('--message','-m','message_template',
              type=click.Path(exists=True),
              help="Mako template to populate and output.")
@click.argument("galaxy")
@click.argument("file",type=click.Path(exists=True))
@pass_context
def create_users_from_file(context,galaxy,file,message_template,
                           only_check):
    """
    Create multiple Galaxy users from a file.

    Creates user accounts in GALAXY instance from contents of
    FILE, which should be a tab-delimited file with details of
    a new user on each line; the columns should be 'email',
    'password', and optionally 'public_name'.

    (If the 'public_name' is missing then it will be generated
    automatically from the leading part of the email.)
    """
    # Check message template is a .mako file
    if message_template:
        if not message_template.endswith(".mako"):
            logger.critical("Message template '%s' is not a .mako file"
                            % message_template)
            return 1
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Create users
    return users.create_batch_of_users(gi,file,
                                       only_check=only_check,
                                       mako_template=message_template)

@nebulizer.command()
@click.option('--name',metavar='NAME',
              help="list only tools matching NAME. Can include "
              "glob-style wild-cards.")
@click.option('--installed','installed_only',is_flag=True,
              help="only list tools that have been installed from "
              "a toolshed (default is to list all tools).")
@click.argument("galaxy")
@pass_context
def list_tools(context,galaxy,name,installed_only):
    """
    List tools in Galaxy instance.

    Prints details of tools available in GALAXY instance,
    including: tool name, version, tool panel section, and
    toolshed repository and revision changeset.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # List tools
    tools.list_tools(gi,name=name,installed_only=installed_only)

@nebulizer.command()
@click.option('--name',metavar='NAME',
              help="only list tool repositories matching NAME. Can "
              "include glob-style wild-cards.")
@click.option('--toolshed',metavar='TOOLSHED',
              help="only list repositories installed from toolshed "
              "matching TOOLSHED. Can include glob-style wild-cards.")
@click.option('--owner',metavar='OWNER',
              help="only list repositories from matching OWNER. "
              "Can include glob-style wild-cards.")
@click.option('--list-tools',is_flag=True,
              help="also list the tools associated with each "
              "installed repository revision changeset.")
@click.option('--updateable',is_flag=True,
              help="only show repositories with uninstalled updates "
              "or upgrades.")
@click.argument("galaxy")
@pass_context
def list_installed_tools(context,galaxy,name,toolshed,owner,list_tools,
                         updateable):
    """
    List installed tool repositories.

    Prints details of installed tool repositories in GALAXY
    instance.

    For each installed repository the details include: repository
    name, toolshed, owner, revision id and changeset, and
    installation status.

    Repository details are also preceeded by a single-character
    'status' indicator ('D' = deprecated; '^' = newer revision
    installed; 'u' = update available but not installed; 'U' =
    upgrade available but not installed; '*' = latest revision).

    If the --list-tools option is specified then additionally
    after each repository the tools associated with the repository
    will be listed along with their descriptions and versions.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # List repositories
    tools.list_installed_repositories(gi,name=name,
                                      toolshed=toolshed,
                                      owner=owner,
                                      list_tools=list_tools,
                                      only_updateable=updateable)

@nebulizer.command()
@click.option('--name',metavar='NAME',
              help="only list tool panel sections where name or "
              "id match NAME. Can include glob-style wild-cards.")
@click.option('--list-tools',is_flag=True,
              help="also list the associated tools for each "
              "section")
@click.argument("galaxy")
@pass_context
def list_tool_panel(context,galaxy,name,list_tools):
    """
    List tool panel contents.

    Prints details of tool panel sections including the
    displayed text and the internal section id, and any
    tools available outside of any section.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # List tool panel contents
    tools.list_tool_panel(gi,name=name,
                          list_tools=list_tools)

@nebulizer.command()
@click.option('--tool-panel-section',
              help="tool panel section name or id to install "
              "the tool under; if the section doesn't exist "
              "then it will be created. If this option is "
              "omitted then the tool will be installed at the "
              "top-level i.e. not in any section.")
@click.option('--timeout',metavar='TIMEOUT',default=600,
              help="wait up to TIMEOUT seconds for tool installations"
              "to complete (default is 600).")
@click.option('--no-wait',is_flag=True,
              help="don't wait for lengthy tool installations to "
              "complete.")
@click.argument("galaxy")
@click.argument("toolshed")
@click.argument("owner")
@click.argument("repository")
@click.argument("revision",required=False)
@pass_context
def install_tool(context,galaxy,toolshed,owner,repository,
                 revision,tool_panel_section,timeout,no_wait):
    """
    Install tool from toolshed.

    Installs the specified tool from REPOSITORY owned by
    OWNER in TOOLSHED, into GALAXY.

    Optionally also specify a changeset REVISION; if no
    revision is specified and no version of the tool is
    already installed then will attempt to install the
    latest revision (otherwise will skip installation).
    Installation will fail if the specified revision is
    not installable, or if no installable revisions are
    found.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Install tool
    return tools.install_tool(
        gi,toolshed,repository,owner,revision=revision,
        tool_panel_section=tool_panel_section,
        timeout=timeout,no_wait=no_wait)

@nebulizer.command()
@click.option('--name',metavar='NAME',
              help="only list tool repositories matching NAME. Can "
              "include glob-style wild-cards.")
@click.option('--toolshed',metavar='TOOLSHED',
              help="only list repositories installed from toolshed "
              "matching TOOLSHED. Can include glob-style wild-cards.")
@click.option('--owner',metavar='OWNER',
              help="only list repositories from matching OWNER. "
              "Can include glob-style wild-cards.")
@click.option('--updateable',is_flag=True,
              help="only show repositories with uninstalled updates "
              "or upgrades.")
@click.argument("galaxy")
@pass_context
def list_repositories(context,galaxy,name,toolshed,owner,updateable):
    """
    List installed tool repos for (re)install.

    Prints details of installed tool repositories in GALAXY
    instance in a format suitable for input into the
    'install_repositories' command.

    The output is a set of tab-delimited values, with each line
    consisting of:

    TOOLSHED|OWNER|REPOSITORY|CHANGESET|TOOL_PANEL_SECTION

    TOOL_PANEL_SECTION will be empty if the repository was
    installed outside of any section in the tool panel.

    The repositories are ordered according to their position
    in the tool panel. Note that non-package and
    non-data-manager repositories which cannot be located
    within the tool panel will not be listed.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # List repositories
    tools.list_installed_repositories(gi,name=name,
                                      toolshed=toolshed,
                                      owner=owner,
                                      only_updateable=updateable,
                                      tsv=True)

@nebulizer.command()
@click.option('--timeout',metavar='TIMEOUT',default=600,
              help="wait up to TIMEOUT seconds for tool installations"
              "to complete (default is 600).")
@click.option('--no-wait',is_flag=True,
              help="don't wait for lengthy tool installations to "
              "complete.")
@click.argument("galaxy")
@click.argument("file",type=click.File('r'))
@pass_context
def install_repositories(context,galaxy,file,timeout,no_wait):
    """
    Install tool repositories listed in a file.

    Installs the tools specified in FILE into GALAXY.

    FILE should be a tab-delimited file with the columns:

    TOOLSHED|OWNER|REPOSITORY|REVISON|SECTION

    If the REVISION field is blank then nebulizer will
    attempt to install the latest revision; if the
    SECTION field is blank then the tool will be
    installed at the top level of the tool panel (i.e.
    not in any section).
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Keep a list of failed tool installs
    failed_install = []
    # Install tools
    for line in file:
        if line.startswith('#'):
            continue
        print line.rstrip('\n')
        line = line.rstrip('\n').split('\t')
        try:
            toolshed,owner,repository = line[:3]
        except ValueError:
            logger.critical("Couldn't parse line")
            return 1
        try:
            revision = line[3]
            if not revision:
                revision = None
        except KeyError:
            revision = None
        try:
            tool_panel_section = line[4]
            if not tool_panel_section:
                tool_panel_section = None
        except KeyError:
            tool_panel_section = None
        status = tools.install_tool(gi,
                                    toolshed,repository,owner,
                                    revision=revision,
                                    tool_panel_section=tool_panel_section,
                                    timeout=timeout,no_wait=no_wait)
        if status != tools.TOOL_INSTALL_OK:
            failed_install.append(line)
    # List any failed tool installations
    if failed_install:
        logger.error("Some requested tool repositories couldn't be "
                     "installed")
        for repo in failed_install:
            click.echo('\t'.join(repo))
        return 1
    # Looks like everything worked
    return 0

@nebulizer.command()
@click.option('--timeout',metavar='TIMEOUT',default=600,
              help="wait up to TIMEOUT seconds for tool installations"
              "to complete (default is 600).")
@click.option('--no-wait',is_flag=True,
              help="don't wait for lengthy tool installations to "
              "complete.")
@click.argument("galaxy")
@click.argument("toolshed")
@click.argument("owner")
@click.argument("repository")
@pass_context
def update_tool(context,galaxy,toolshed,owner,repository,
                timeout,no_wait):
    """
    Update tool installed from toolshed.

    Updates the specified tool from REPOSITORY owned by
    OWNER in TOOLSHED, to the latest revision in GALAXY.

    The tool must already be present in GALAXY and a newer
    changeset revision must be available. The update will
    be installed into the same tool panel section as the
    original tool.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Install tool
    return tools.update_tool(gi,toolshed,repository,owner,
                             timeout=timeout,no_wait=no_wait)

@nebulizer.command()
@click.option('-l','long_listing',is_flag=True,
              help="use a long listing format that includes "
              "ids, descriptions, file sizes, dbkeys and paths)")
@click.argument("galaxy")
@click.argument("path",required=False)
@pass_context
def list_libraries(context,galaxy,path,long_listing):
    """
    List data libraries and contents.

    Prints details of the data library and folders
    specified by PATH, in GALAXY instance.

    PATH should be of the form
    'data_library[/folder[/subfolder[...]]]'
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # List folders in data library
    if path:
        libraries.list_library_contents(gi,path,
                                        long_listing_format=
                                        long_listing)
    else:
        libraries.list_data_libraries(gi)

@nebulizer.command()
@click.option('-d','--description',
              help="description of the new library")
@click.option('-s','--synopsis',
              help="synopsis text for the new library")
@click.argument("galaxy")
@click.argument("name")
@pass_context
def create_library(context,galaxy,name,description,synopsis):
    """
    Create new data library.

    Makes a new data library NAME in GALAXY. A library
    with the same name must not already.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Create new data library
    libraries.create_library(gi,name,
                             description=description,
                             synopsis=synopsis)

@nebulizer.command()
@click.option('-d','--description',
              help="description of the new folder")
@click.argument("galaxy")
@click.argument("path")
@pass_context
def create_library_folder(context,galaxy,path,description):
    """
    Create new folder in a data library.

    Makes a new folder or folder tree within an existing data
    library in GALAXY.

    The new folder(s) are specified by PATH, which should be
    of the form 'data_library[/folder[/subfolder[...]]]'.
    Although the data library must already exist, PATH must
    not address an existing folder.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Create new folder
    libraries.create_folder(gi,path,description=description)

@nebulizer.command()
@click.option('--file-type',default='auto',
              help="Galaxy data type to assign the files to "
              "(default is 'auto'). Must be a valid Galaxy "
              "data type. If not 'auto' then all files will "
              "be assigned the same type.")
@click.option('--dbkey',default='?',
              help="dbkey to assign to files (default is '?')")
@click.option('--server','from_server',is_flag=True,
              help="upload files from the Galaxy server file "
              "system (default is to upload files from local "
              "system)")
@click.option('--link',is_flag=True,
              help="create symlinks to files on server (only "
              "valid if used with --server; default is to copy "
              "files into Galaxy)")
@click.argument("galaxy")
@click.argument("dest")
@click.argument("file",nargs=-1)
@pass_context
def add_library_datasets(context,galaxy,dest,file,file_type,
                         dbkey,from_server,link):
    """
    Add datasets to a data library.

    Uploads one or more FILEs as new datasets in the
    data library folder DEST in GALAXY.

    DEST should be a path to a data library or library
    folder of the form
    'data_library[/folder[/subfolder[...]]]'. The library
    and folder must already exist.
    """
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    # Add the datasets
    libraries.add_library_datasets(gi,dest,file,
                                   from_server=from_server,
                                   link_only=link,
                                   file_type=file_type,
                                   dbkey=dbkey)

@nebulizer.command()
@click.option('-c','--count',metavar='COUNT',default=0,
              help="if set then stop after sending COUNT requests "
              "(default is to send requests forever).")
@click.option('-i','--interval',metavar='INTERVAL',default=5,
              help="set the interval between sending requests in "
              "seconds (default is 5 seconds).")
@click.argument("galaxy")
@pass_context
def ping(context,galaxy,count,interval=5):
    """
    'Ping' a Galaxy instance.

    Sends a request to GALAXY and reports the status of the
    response and the time taken.
    """
    try:
        galaxy_url,_ = Credentials().fetch_key(galaxy)
    except KeyError:
        galaxy_url = galaxy
    click.echo("PING %s" % galaxy_url)
    nrequests = 0
    while True:
        try:
            # Get a Galaxy instance
            gi = context.galaxy_instance(galaxy_url)
            if gi is None:
                click.echo("%s: failed to connect" % galaxy_url)
                return 1
            else:
                status_code,response_time = ping_galaxy_instance(gi)
                if status_code != 0:
                    msg = "failed (error code %s)" % status[0]
                else:
                    msg = "ok"
                click.echo("%s: status = %s time = %.3f (ms)" %
                           (galaxy_url,msg,response_time*1000.0))
            # Deal with count limit, if set
            if count != 0:
                nrequests += 1
                if nrequests >= count:
                    break
            # Wait before sending next request
            time.sleep(interval)
        except KeyboardInterrupt:
            break
    return status_code

@nebulizer.command()
@click.argument("galaxy")
@pass_context
def whoami(context,galaxy,):
    """
    Print user details associated with API key.
    """
    logger.debug("Debugging mode")
    # Get a Galaxy instance
    gi = context.galaxy_instance(galaxy)
    if gi is None:
        logger.critical("Failed to connect to Galaxy instance")
        return 1
    user = get_current_user(gi)
    if user is None:
        logger.warning("No associated user for this API key")
    else:
        click.echo("%s" % user['email'])
